% Poly-Prolog, RH, 29.09.1993

% Poly-Prolog ist im wesentlichen das ALF-System von Michael Hanus.
% Spezielle Eigenschaften:
%	- Polymorphes Typsystem a la Hanus.
%	- Optimierung der polymorphen Unifikation (parametrisierte Form).
%	- Unterscheidung zwischen Funktionen und Konstruktoren.
%	- Innermost Narrowing durch Entschachtelung von Funktionsaufrufen.
%	- Funktionen k"onnen alternativ relational geschrieben werden.
%	- Listenbeschreibungen (mit und ohne Duplikate).
%	- Rumpf einer Klausel bzw. einer Listenbeschreibung darf eine
%	  logische Formel sein.
%	- Einfaches Modulkonzept.
%	- Viele vordefinierte Prolog-Pr"adikate und Junktoren verf"ugbar.
%	- Esacpe nach Prolog mit `{_}'.
% W"unschenswerte Verbesserungen:
%	- Sprache: die Formeln m"ussen bis dato den Variablenkonventionen
%	  gen"ugen ([(X,[Y <- p(X,Y)]) <- Y^p(X,Y)] ist unzul"assig).
%	- Implementierung: bei den neu generierten Pr"adikaten fehlen
%	  noch die Typparameter.
%	- Implementierung: die "ubersetzten Klauseln enthalten noch
%	  Aufrufe von `='.
%	- Implementierung: bei der Antwortsubstitution werden keine
%	  Typvariablen ausgeben.
%	- Implementierung: bei der Antwortsubstitution *nicht* die
%	  parametrisierte Form ausgeben.
%	- Implementierung: Erkennung obsoleter Argumentpositionen.
%	- Implementierung: `print_message' und `on_exception', 
%	  `raise_exception'.
%	- Implementierung: Modusdeklarationen sollten ber"ucksichtigt
%	  werden.
%	- Implementierung: Funktionen und Konstruktoren gleichen Namens
%	  werden nicht abgefangen.
%	- Implementierung: `<-', `<=' und `{_}' sind reservierte
%	  Symbole, sollten vom Benutzer nicht f"ur Funktionen oder
%	  Pr"adikate verwendet werden.
%	- Implementierung: wird `<=' oder `<-' im Rumpf einer Klausel
%	  verwendet, ger"at der "Ubersetzer in eine Endlosschleife.
%       - Implementierung: Die Integritaetspruefung wird zur Zeit nur 
%	  naiv durchgefuehrt, d.h. Auswertung der Anfrage yes <- ic.
%	- Implementierung: Die Ausfuehrung von Aenderungsanweisungen ist
%	  nicht moeglich, da die Transformation von Aenderungsanweisungen 
%	  und Aenderungsprozeduren in das interne Format noch implementiert 
%	  werden muss.

:- module(ddbs, [
	ddbs/0
   ]).

%-------------------------------- Bibliotheken ---------------------------------

% restore the status of operators for compiling the libraries

reset_op :-
        current_op(_, Type, Name),
        op(0, Type, Name),
        fail.
reset_op :- 
	op(1200, xfx, [ :-, --> ]),
	op(1200,  fx, [ :-, ?- ]),
	op(1150,  fx, [ dynamic, multifile, meta_predicate, fx, init,
                         volatile, mode, public ]),
	op(1100, xfy, [ ; ]),
	op(1050, xfy, [ -> ]),
	op(1000, xfy, [ ',' ]),
	op( 900,  fy, [ \+, spy, nospy ]),
	op( 700, xfx, [ =, is, =.., ==, \==, @<, @>, @=<, @>=,
        	                            =:=, =\=, <, >, =<, >= ]),
	op( 600, xfy, [ : ]),
	op( 500, yfx, [ +, -, \/, /\ ]),
	op( 500,  fx, [ +, - ]),
	op( 400, yfx, [ /, div, //, *, <<, >> ]),
	op( 300, xfx, [ mod ]),
	op( 200, xfy, [ ^ ]).

:- reset_op.

:- use_module(library(unify), [
	unify/2
   ]),
   use_module(library(lineio), [
	get_line/1
   ]),
   use_module(library(strings), [
	string_append/3,
	concat_atom/2,
	gensym/2
   ]),
   use_module(library(sets), [
        union/2
   ]),
   use_module(library(lists), [
        delete/3
   ]),
   use_module(library(same_functor), [
	same_functor/2,
	same_functor/4
   ]),
   use_module(library(samsort), [
	samsort/2
   ]),
   use_module(library(print_chars)).    

%--------------------------------- Operatoren ----------------------------------

:- op(1150,  fx, [ compile, 
                     include, quantifier, connector, system, 
                     type, data, cons, func, pred, dynamic, mode, 
                     ic, proc ]).                                

:- op(1150, xfy, [ ::= ]).
:- op( 975, xfx, [ <-, <= ]).
:- op( 600, yfx, [ : ]).

%--------------------------- Dynamische Pr"adikate -----------------------------

:- dynamic quan_decl/4.
:- dynamic conn_decl/4.
:- dynamic type_decl/4.
:- dynamic cons_decl/4.
:- dynamic func_decl/5.
:- dynamic pred_decl/3.
:- dynamic proc_decl/3.
:- dynamic loaded/1.
:- dynamic flag/1.
:- dynamic ddbs_error/0.

retract_dynamic :-
	retractall(quan_decl(_,_,_,_)),
	retractall(conn_decl(_,_,_,_)),
	retractall(type_decl(_,_,_,_)),
	retractall(cons_decl(_,_,_,_)),
	retractall(func_decl(_,_,_,_,_)),
	retractall(pred_decl(_,_,_)),
        retractall(proc_decl(_,_,_)),
	retractall(loaded(_)),
	retractall(flag(_)),
	retractall(ddbs_error).

retract_clauses :-
	(   bagof(user:Functor/Arity, Name^Pred^Skel^
                  (pred_decl(Name, Pred, Skel),
		   \+ protected(user:Skel),
                   functor(Skel, Functor, Arity)), ProcSpecs) ->
		abolish(ProcSpecs)
	;	true
	).

protected(Skel) :-
	predicate_property(Skel, built_in);
	predicate_property(Skel, imported_from(_)).

%--------------------------- Benutzerschnittstelle -----------------------------

ddbs :-
	prompt(_, '<DDBS>: '),			% initialize
	nl, 
	rjustify(80, "Into this world we're born"),
	rjustify(80, "into this world we're thrown"),
	rjustify(80, "The Doors"), 
	nl,
	no_style_check(all),
	nofileerrors,
	retract_clauses,
	retract_dynamic,
	compile_file(stdenv),
	on_exception(Exception, cycle, handle(Exception)).

cycle :-
	read_term([variable_names(Vars)], Goal),
	cycle(Goal, Vars).

cycle(end_of_file, _) :- !,			% terminate
	style_check(all),
	nl, 
	rjustify(80,"This is the end"),
	rjustify(80,"my only friend the end"),
	rjustify(80,"The Doors"), 
	nl.
cycle(Goal, Vars) :-
	execute(Goal, Vars), !,			% `execute' should not fail!
	nl,
        cycle.

execute((compile File), _) :- !,		% overwrite `compile'
	retract_clauses,
	retract_dynamic,
        compile_file(stdenv),
        (   compile_file(File) 
        ;   true
	).
execute([File], _) :- !,
	execute((compile File), _).
execute(listing, _) :- !,
	listing(quan_decl/4),
	listing(conn_decl/4),
	listing(type_decl/4),
	listing(cons_decl/4),
	listing(func_decl/5),
        listing(pred_decl/3),
        listing(proc_decl/3).
execute(count, _)  :- !,
	assert(flag(count)).
execute(nocount, _)  :- !,
	retractall(flag(count)).
execute(icc, _) :- !,
	write('Only naive integrity constraint checking'),nl,
        (   user:(ic) -> 
	    	nl,write('Integrity constraints satisfied'),nl,nl
	;   	nl,write('Integrity constraints violated'),nl,nl
	).
execute(Update, _) :-   % we ignore update statements
        nonvar(Update),
        Update = {_},
        !.
%       (   transform_update(Update,TUpdate) ->
%           	evaluate_update(TUpdate,Del,Ins),
%               print_update(Del,'Deleted:  '),nl,
%               print_update(Ins,'Inserted: ')
%	).
execute(<=(Body), _) :-
	free_variables(Body, [], Frees),
	list_to_tuple(Frees, Tuple),
	execute((Tuple<= Body), _).
execute((Head <= Body), _) :-
	(   assert_query(('$query'(Head) :- Body)) ->
		(   bagof((T,X), user:'$query'(X,T), Bag) ->
			samsort(Bag, Solutions),
			print_list(Solutions)
		;	format('~nno~n', [])
		)
	;	true
	).
execute(<-(Body), _) :-
	free_variables(Body, [], Frees),
	list_to_tuple(Frees, Tuple),
	execute((Tuple<- Body), _).
execute((Head <- Body), _) :-
	(   assert_query(('$query'(Head) :- Body)) ->
		(   setof((T,X), user:'$query'(X,T), Solutions) ->
			print_list(Solutions)
		;	format('~nno~n', [])
		)
	;	true
	).
execute(Goal, Vars) :-
	free_variables(Goal, [], Frees),
	list_to_tuple(Frees, Tuple),
	(   assert_query(('$query'(Tuple) :- Goal)) ->
		remove_bound(Vars, Frees, FreeVars),
		execute_goal(Tuple, FreeVars)
	;	true
	).

all_solutions((Head <- Body), Solutions) :-
	assert_query(('$query'(Head) :- Body)) ->
        (   setof(X, T^(user:'$query'(X,T)), Solutions) -> true  
	;	Solutions = []
	).

execute_goal(Tuple, FreeVars) :-
	statistics(runtime,[T0|_]),
	user:'$query'(Tuple,_),
	statistics(runtime,[T1|_]),
	T is T1-T0,
	(   flag(count) ->		% print statistics
		format('~nruntime: ~3d sec.',[T])
	;	true
	),
	numbervars(FreeVars, 0, _),
	(   FreeVars = [] ->
		format('~nyes', [])
	;	print_bindings(FreeVars),	% document success
		get_line([])			% wants the user more solutions?
	).
execute_goal(_, _) :-
	format('~nno', []).		% document failure

handle(Message) :-
	print_message(error, Message),
	on_exception(Exception, cycle, handle(Exception)).

list_to_tuple([], yes). 
list_to_tuple([A], A) :- !.
list_to_tuple([A|X], (A,Y)) :-	% Konstruktor `,' mu"s deklariert sein
	list_to_tuple(X, Y).

remove_bound([], _, []).
remove_bound([Name=Var|Vars], Frees, FreeVars) :-
	(   list_is_free_of(Frees, Var) ->
		FreeVars=FreeVars0
	;	FreeVars=[Name=Var|FreeVars0]
	),
	remove_bound(Vars, Frees, FreeVars0).

print_list([]).
print_list(Solutions) :-
	Solutions = [(Type,_)|_],
	format('~nType: ~w~n~n', [Type]),
        print_list(Type, Solutions).

print_list(_Type, []).
print_list(Type, Solutions) :-
	(   Solutions = [(Type,Elem)|Rest] ->
		print(Elem), nl,
		print_list(Type,Rest)
	;	print_list(Solutions)
	).

print_bindings([]) :-
	format('~nmore solutions? ', []).
print_bindings([Var=Term|Bindings]) :-
	format('~n~w = ',[Var]),
	print(Term),
	print_bindings(Bindings).

%-------------------------------- "Ubersetzer ----------------------------------

compile_file(File) :-
        (   absolute_file_name(File, [extensions(['ddb','']),
                                      access(read)], Source),
                string_append(Prefix,'.ddb',Source),
                string_append(Prefix,'.pl',Target) ->
		(   loaded(Source) ->
			format('% ddb file ~a already loaded~n',Source);
	                (   absolute_file_name(Target, [extensions(['']),
        	                                        access(write)], _) ->
                	        format('% compiling ddb file ~a~n',Source),
	                        open(Source,read,SourceStream),
	                        open(Target,write,TargetStream),
        	                translate_stream(SourceStream,TargetStream),
                	        compile(user:Target),
				assert(loaded(Source))
        	        ;       format('*** cannot write ~a~n',[Target]),
				fail
	                )
		)
        ;       format('*** cannot read ~a (or wrong extension)~n',[File]),
		fail
        ).

translate_stream(SourceStream, TargetStream) :-
	line_count(SourceStream, From),
	read(SourceStream, Term),
	line_count(SourceStream, To),
	translate_stream(Term, SourceStream, TargetStream, From, To).

translate_stream(end_of_file, SourceStream, TargetStream, _, _) :- !,
	close(SourceStream),			% terminate
        translate_ic_top_clause(TargetStream),        
	close(TargetStream).
translate_stream(Term, SourceStream, TargetStream, From, To) :-
	translate(Term, TargetStream),
	(   ddbs_error ->
		format('*** between lines ~d and ~d~n', [From,To]),
		retractall(ddbs_error)
	;	true
	),
	translate_stream(SourceStream, TargetStream).

%----------------------- "Ubersetzung von Deklarationen ------------------------

translate((include File), TargetStream) :- !,
	translate_decl((include), File, TargetStream).
translate((quantifier Decl), TargetStream) :- !,
	translate_decl((quantifier), Decl, TargetStream).
translate((connector Decl), TargetStream) :- !,
	translate_decl((connector), Decl, TargetStream).
translate((system Decl), TargetStream) :- !,
	translate_decl((system), Decl, TargetStream).
translate((type Decl), TargetStream) :- !,
	translate_decl((type), Decl, TargetStream).
translate((data Decl), TargetStream) :- !,
	translate_decl((data), Decl, TargetStream).
translate((cons Decl), TargetStream) :- !,
	translate_decl((cons), Decl, TargetStream).
translate((func Decl), TargetStream) :- !,
	translate_decl((func), Decl, TargetStream).
translate((pred Decl), TargetStream) :- !,
	translate_decl((pred), Decl, TargetStream).
translate((dynamic Decl), TargetStream) :- !,
	translate_decl((dynamic), Decl, TargetStream).
translate((mode _), _) :- !.			% we ignore `mode'-declarations
translate((proc Decl), TargetStream) :- !,
        translate_decl((proc), Decl, TargetStream).
translate((ic IC), TargetStream) :- !,
        translate_ic(IC, TargetStream).
translate((_Head ::= _Body), _TargetStream) :- !. % we ignore update procedures
translate(op(P,A,N), TargetStream) :- !,
	op(P, A, N),
	portray_clause(TargetStream, (:- op(P,A,N))).
translate((:- Goal), TargetStream) :- !,
	portray_clause(TargetStream, (:- Goal)).
translate(Formula, TargetStream) :-		% everything else is a clause
	reduce_formula(Formula, Simple),
	(   Simple = (Head :- Body) -> true
	;	Head = Formula,
		Body = true
	),
	annotate((Head :- Body), _X, Clauses),
	forall(member(Clause, Clauses),
	       portray_clause(TargetStream, Clause)).

translate_decl(What, (Decl1,Decl2), TargetStream) :- !,
	translate_decl(What, Decl1, TargetStream),
	translate_decl(What, Decl2, TargetStream).
translate_decl((include), File, _) :- !,
	compile_file(File).
translate_decl((quantifier), QuantCons/Arity, _) :- !,
	functor(Quant, QuantCons, Arity),
	same_functor(Quant, CopyQuant),
	add_declaration(quan_decl(Quant,CopyQuant,CopyQuant,Arity)).
translate_decl((quantifier), Quant=Abbr, _) :- !,
	reduce_formula(Abbr, RedAbbr),
	copy_term(Quant=RedAbbr,CopyQuant=CopyAbbr),
	functor(Quant, _, Arity),
	add_declaration(quan_decl(Quant,CopyQuant,CopyAbbr,Arity)).
translate_decl((connector), ConnCons/Arity, _) :- !,
	functor(Conn, ConnCons, Arity),
	same_functor(Conn, CopyConn),
	add_declaration(conn_decl(Conn,CopyConn,CopyConn,Arity)).
translate_decl((connector), Conn=Abbr, _) :- !,
	reduce_formula(Abbr, RedAbbr),
	copy_term(Conn=RedAbbr,CopyConn=CopyAbbr),
	functor(Conn, _, Arity),
	add_declaration(conn_decl(Conn,CopyConn,CopyAbbr,Arity)).
translate_decl((type), TypeCons/Arity, _) :- !,
	functor(Type, TypeCons, Arity),
	same_functor(Type, CopyType),
	add_declaration(type_decl(Type,CopyType,CopyType,Arity)).
translate_decl((type), Type=Abbr, _) :- !,
	reduce_type(Abbr, RedAbbr),
	copy_term(Type=RedAbbr,CopyType=CopyAbbr),
	functor(Type, _, Arity),
	add_declaration(type_decl(Type,CopyType,CopyAbbr,Arity)).
translate_decl((system), Func:Type, _) :- !,
	same_functor(Func, RedFunc, Functor, Arity),
	reduce_type_arg(Arity, Func, RedFunc),
	reduce_type(Type, RedType),
	ext_parameters(RedFunc,[RedType],Pred),
	same_functor(Pred, Skel, _, Arity1),
	arg(Arity1, Skel, Result),
% Beachte: Mehrfachdeklarationen werden nicht abgefangen.
	assertz(func_decl(Functor,RedFunc,RedType,Skel,Result)),
	assertz(pred_decl(Functor,Pred,Skel)).
translate_decl((system), Pred, _) :- !,
	same_functor(Pred, RedPred, Functor, Arity),
	reduce_type_arg(Arity, Pred, RedPred),
	same_functor(RedPred, Skel),
% Beachte: Mehrfachdeklarationen werden nicht abgefangen.
	assertz(pred_decl(Functor,RedPred,Skel)).
translate_decl((data), Type={Constrs}, _) :- !,
	functor(Type, Functor, Arity),
	translate_decl((type), Functor/Arity, _),
	translate_data(Constrs, Type, _).
translate_decl((cons), Cons:Type, _) :- !,
	same_functor(Cons, RedCons, Functor, Arity),
	reduce_type_arg(Arity, Cons, RedCons),
	reduce_type(Type, RedType),
	parametric_form(RedCons, RedType, ParaForm),
	add_declaration(cons_decl(Functor,RedCons,RedType,ParaForm)).
translate_decl((func), Func:Type, _) :- !,
	same_functor(Func, RedFunc, Functor, Arity),
	reduce_type_arg(Arity, Func, RedFunc),
	reduce_type(Type, RedType),
	ext_parameters(RedFunc,[RedType],Pred),
	parametric_form(Pred, bool, FlatForm),
	Arity1 is Arity+1,
	arg(Arity1, FlatForm, Result),
	add_declaration(func_decl(Functor,RedFunc,RedType,FlatForm,Result)),
	add_declaration(pred_decl(Functor,Pred,FlatForm)).
translate_decl((pred), Pred, _) :- !,
	same_functor(Pred, RedPred, Functor, Arity),
	reduce_type_arg(Arity, Pred, RedPred),
	parametric_form(RedPred, bool, ParaForm),
	add_declaration(pred_decl(Functor,RedPred,ParaForm)).
translate_decl((dynamic), Pred, TargetStream) :- !,
	same_functor(Pred, RedPred, Functor, Arity),
	reduce_type_arg(Arity, Pred, RedPred),
	parametric_form(RedPred, bool, ParaForm),
	add_declaration(pred_decl(Functor,RedPred,ParaForm)),
	functor(ParaForm, _, Arity1),
	portray_clause(TargetStream, (:- dynamic Functor/Arity1)).
translate_decl((proc), SProc, _) :- !,
        reduce_sign(SProc,Proc,Sign),
        same_functor(Proc, RedProc, Functor, Arity),
        reduce_type_arg(Arity, Proc, RedProc),
        parametric_form(RedProc, bool, ParaForm),
        reduce_sign(SRedProc,RedProc,Sign),
        reduce_sign(SParaForm,ParaForm,Sign),
        reduce_sign(SFunctor,Functor,Sign),
        add_declaration(proc_decl(SFunctor,SRedProc,SParaForm)).
translate_decl(What, Decl, _) :-
	ddbs_error('*** illegal declaration: ~w ~w~n', [What,Decl]).

translate_data((Constr1,Constr2), Type, _) :- !,
	translate_data(Constr1, Type, _),
	translate_data(Constr2, Type, _).
translate_data(Constr, Type, _) :-
	translate_decl((cons), Constr:Type, _).

translate_ic(IC, TargetStream) :-
        gensym(ic, ICSymbol),
        translate((pred ICSymbol), TargetStream),
        translate((ICSymbol <- IC), TargetStream).

translate_ic_top_clause(TargetStream) :-
        (   \+ pred_decl((ic),_,_) ->
            translate((pred (ic)), TargetStream)
        ;   true
        ),
        (   setof(IC,[X,Y,Z,S]^(pred_decl(IC,X,Y),
                                name(IC,S),
                                append("ic",Z,S)),
                  ICList) ->
            delete(ICList,(ic),ICList2),
            list_to_conjunction(ICList2,ICBody)
        ;   ICBody = true
        ),
        translate((ic :- ICBody), TargetStream).

reduce_sign(+Proc,Proc,+).
reduce_sign(-Proc,Proc,-).
reduce_sign(Proc,Proc,no).

reduce_type(Type, RedType) :-
	(   var(Type) ->
		Type = RedType
	;	(   type_decl(Type, Copy, RedType, Arity) ->
			reduce_type_arg(Arity, Type, Copy)
		;	same_functor(Type, RedType, Functor, Arity),
			ddbs_error('*** ~w not declared~n', [Functor/Arity]),
			% recovery: assume type declaration `type Functor/Arity'
			reduce_type_arg(Arity, Type, RedType)
		)
	).

reduce_type_arg(N, Type, Copy) :-
	(   N =:= 0 ->
		true
	;	arg(N, Type, TypeArg),
		arg(N, Copy, CopyArg),
		reduce_type(TypeArg, CopyArg),
		N1 is N-1,
		reduce_type_arg(N1, Type, Copy)
	).

reduce_formula(Form, RedForm) :-
	(   var(Form) ->			% meta-variable
		Form = RedForm
	;   quan_decl(Form, Copy, RedForm, Arity) ->
		arg(1, Form, Variables),	% copy quantified variables
		arg(1, Copy, Variables),
		reduce_formula_arg(2, Arity, Form, Copy)
	;   conn_decl(Form, Copy, RedForm, Arity) ->
		reduce_formula_arg(1, Arity, Form, Copy)
	;	Form = RedForm			% atomic formula
	).

reduce_formula_arg(N, Arity, Form, Copy) :-
	(   N > Arity ->
		true
	;	arg(N, Form, FormArg),
		arg(N, Copy, CopyArg),
		reduce_formula(FormArg, CopyArg),
		N1 is N+1,
		reduce_formula_arg(N1, Arity, Form, Copy)
	).

ext_parameters(Atom,Parameters,ExtAtom) :-
	Atom =.. List,
	append(List, Parameters, ExtList),
	ExtAtom =.. ExtList.

parametric_form(Term, Type, ParaForm) :-
	free_variables(Type, [], TypeVars),
	free_variables(Term, TypeVars, ParaVars),
	same_functor(Term, Skel),
	ext_parameters(Skel, ParaVars, ParaForm).

add_declaration(NewDecl) :-
	same_functor(NewDecl, OldDecl),
	arg(1, NewDecl, Functor),
	arg(1, OldDecl, Functor),
	(   call(OldDecl) ->
		ddbs_error('*** ~w already declared~n', [Functor])
	;	assertz(NewDecl)
	).

portray_clause(Stream, Clause) :-
	current_output(OldStream),
	set_output(Stream),
	portray_clause(Clause),
	set_output(OldStream).

%------------------------- "Ubersetzung von Klauseln ---------------------------

assert_query(Goal) :-
	retractall(user:'$query'(_,_)),
	reduce_formula(Goal, Simple),
	annotate(Simple, _X, Clauses), 
	(   ddbs_error ->
		retractall(ddbs_error),
		fail
	;	forall(member(Clause, Clauses),
		       assertz(user:Clause))
	).

annotate((LHS=RHS:-Body), X, Normalform) :- !,
	(   nonvar(LHS),
	    functor(LHS, Functor, _),
	    func_decl(Functor, _, _, _, _) ->
		ext_parameters(LHS,[RHS],Head),
		annotate_atom(Head, AnnHead, Atoms, X, Clauses1),
		annotate_formula(Body, AnnBody, X, Clauses2),
					% Beachte den `!'
	        disjunctive_nf(and(AnnBody,and(!,Atoms)), DNF, yes),
	        make_clauses(AnnHead, DNF, Clauses3),
        	append(Clauses1, Clauses2, Aux),
	        append(Aux, Clauses3, Normalform)
	;	ddbs_error('*** illegal equation: ~w=... ~n', [LHS])
	), !.
annotate((Head:-Body), X, Normalform) :- !,
	annotate_atom(Head, AnnHead, Atoms, X, Clauses1),
	annotate_formula(Body, AnnBody, X, Clauses2),
        disjunctive_nf(and(AnnBody,Atoms), DNF, yes),
        make_clauses(AnnHead, DNF, Clauses3),
        append(Clauses1, Clauses2, Aux),
        append(Aux, Clauses3, Normalform), !.

annotate_formula(not(ex(Ys,W)), \+ NewAtom, X, Clauses) :- !,
        free_variables(ex(Ys,W), [], Xs),
        gensym('$new', Symbol),
        NewAtom =.. [Symbol|Xs],
	same_functor(NewAtom, Spec, Functor, _),	% generiere Typdekl.
	same_functor(NewAtom, Skel, Functor, _),
	assertz(pred_decl(Functor,Spec,Skel)),	
        annotate((NewAtom :- W), X, Clauses).
annotate_formula(not(and(W1,W2)), V, X, Clauses) :- !,
        annotate_formula(or(not(W1),not(W2)), V, X, Clauses).
annotate_formula(not(or(W1,W2)), V, X, Clauses) :- !, 
        annotate_formula(and(not(W1),not(W2)), V, X, Clauses).
annotate_formula(not(not(W)), V, X, Clauses) :- !, 
        annotate_formula(W, V, X, Clauses).
annotate_formula(ex(_Ys,W), V, X, Clauses) :- !, 
	annotate_formula(W, V, X, Clauses).
annotate_formula(and(W1,W2), and(V1,V2), X, Clauses) :- !, 
        annotate_formula(W1, V1, X, Clauses1), 
        annotate_formula(W2, V2, X, Clauses2), 
        append(Clauses1, Clauses2, Clauses).
annotate_formula(or(W1,W2), or(V1,V2), X, Clauses) :- !, 
        annotate_formula(W1, V1, X, Clauses1), 
        annotate_formula(W2, V2, X, Clauses2), 
        append(Clauses1, Clauses2, Clauses).
annotate_formula(not(Atom), \+ V, X, Clauses) :- !,
	annotate_formula(Atom, W, X, Clauses),
	disjunctive_nf(W, V, no).		% falls Formula `and' enth"alt
% Beachte: `V' kann eine Prolog-Formel sein.
annotate_formula(Form, AnnForm, X, Clauses) :-
	quan_decl(Form, AnnForm, _, Arity), !,
	arg(1, Form, Variables),		% copy quantified variables
	arg(1, AnnForm, Variables),
	annotate_meta_args(2, Arity, Form, AnnForm, X, Clauses).
annotate_formula(Form, AnnForm, X, Clauses) :-
	conn_decl(Form, AnnForm, _, Arity), !,
	annotate_meta_args(1, Arity, Form, AnnForm, X, Clauses).
annotate_formula({Atom}, Atom, _X, []).		% protected Atom
annotate_formula(Atom, Formula, X, Clauses) :-
	annotate_atom(Atom, AnnAtom, Atoms, X, Clauses),
	(   Atoms = true ->
		Formula = AnnAtom
	;	Formula = and(Atoms,AnnAtom)	% Beachte: `and'.
	).

annotate_meta_args(N, Arity, Form, AnnForm, X, Clauses) :- 
        (   N > Arity -> 
		Clauses = []
	;	arg(N, Form, Arg),
		arg(N, AnnForm, AnnArg),
		annotate_formula(Arg, AnnArg, X, Clauses1),
		N1 is N + 1,
		annotate_meta_args(N1, Arity, Form, AnnForm, X, Clauses2),
	        append(Clauses1, Clauses2, Clauses)
	).

annotate_atom(Atom, AnnAtom, Atoms, X, Clauses) :-
	same_functor(Atom, Spec, Functor, Arity),
	(   pred_decl(Functor, Spec, AnnAtom) -> true
	;	ddbs_error('*** ~w not declared~n', [Functor/Arity]),
		same_functor(Atom, AnnAtom, Functor, Arity)
		% recovery: assume most general type
	),
	annotate_args(Arity, Atom, Spec, AnnAtom, X, Clauses, Atoms, true).

annotate_args(N, Term, Spec, AnnTerm, X, Clauses, Atoms0, Atoms) :- 
        (   N =:= 0 -> 
		Clauses = [],
		Atoms0 = Atoms
	;	arg(N, Term, Arg),
		arg(N, Spec, SpecArg), 
		arg(N, AnnTerm, AnnArg), 
		annotate_term(Arg, SpecArg, AnnArg, X, Clauses1, Atoms1, Atoms), 
		N1 is N-1, 
		annotate_args(N1, Term, Spec, AnnTerm, X, Clauses2, Atoms0, Atoms1),
	        append(Clauses1, Clauses2, Clauses)
	).

annotate_term(Var, Type, Var, X, [], Atoms, Atoms) :-
	var(Var), !,			% this cut is necessary
	contains(Var, Spec, X), 
	check(Type, Spec).
annotate_term(Num, Type, Num, _X, [], Atoms, Atoms) :-
	number(Num), !,			% this cut is necessary
	check(Type, num).	
annotate_term(Term:AnnType, Type, AnnTerm, X, Clauses, Atoms0, Atoms) :- !,	% this cut is necessary
	reduce_type(AnnType, RedType),	
	check(Type, RedType), 
	annotate_term(Term, Type, AnnTerm, X, Clauses, Atoms0, Atoms).
annotate_term(Compr, Type, NewVar, X, Clauses, and(Atom,Atoms), Atoms) :-
	comprehension(Compr, Head, Body, OfSymbol), !,
	check(Type, list(A)),
	% neue Listenbeschreibung
	annotate_term(Head, A, AnnHead, X, Clauses1, HeadAtoms, true),
        gensym('$new', Symbol),
        free_variables(Body, [], FreeBody),
        NewAtom =.. [Symbol|FreeBody],
	same_functor(NewAtom, Spec, Functor, _),	% generiere Typdekl.
	same_functor(NewAtom, Skel, Functor, _),
	assertz(pred_decl(Functor,Spec,Skel)),	
	reduce_formula(Body, Simple),
        annotate((NewAtom :- Simple), X, Clauses2),
	disjunctive_nf(and(NewAtom,HeadAtoms), NewBody, no),
	% zu existenzquantifizierende Variablen bestimmen
	free_variables(Head, [], Local),
        free_variables(Body, Local, Global),
	free_variables(AnnHead, [], NewLocal),
        free_variables(NewBody, NewLocal^Global, Bound),
	Atom =.. [OfSymbol,AnnHead,Bound^NewBody,NewVar],
	append(Clauses1, Clauses2, Clauses).
annotate_term(Term, Type, NewVar, X, Clauses, Atoms0, Atoms) :-
	same_functor(Term, Spec, Functor, Arity), 
	func_decl(Functor, Spec, ResultType, Atom, NewVar),
	check(Type, ResultType), 
	annotate_args(Arity, Term, Spec, Atom, X, Clauses, Atoms0, and(Atom,Atoms)).
annotate_term(Term, Type, AnnTerm, X, Clauses, Atoms0, Atoms) :-
	same_functor(Term, Spec, Functor, Arity), 
	(   cons_decl(Functor, Spec, ResultType, AnnTerm) ->
		true
	;	ddbs_error('*** ~w not declared~n', [Functor/Arity]), 
		same_functor(Term, AnnTerm, Functor, Arity)
		% recovery: assume most general type
	), 
	check(Type, ResultType), 
	annotate_args(Arity, Term, Spec, AnnTerm, X, Clauses, Atoms0, Atoms).

make_clauses(Head, (or(true,Bodies)), [Head|Clauses]) :- !,
        make_clauses(Head, Bodies, Clauses).
make_clauses(Head, (or(Body,Bodies)), [(Head :- Body)|Clauses]) :- !,
        make_clauses(Head, Bodies, Clauses).
make_clauses(Head, true, [Head]).
make_clauses(Head, Body, [(Head :- Body)]).

disjunctive_nf(or(A,B), DNF, _Level) :- !,
        disjunctive_nf(A, DNFA, no),
        disjunctive_nf(B, DNFB, no),
	disjunction(DNFA, DNFB, DNF).
disjunctive_nf(and(A,B), DNFAB, Level) :- !,
        disjunctive_nf(A, DNFA, Level),
        disjunctive_nf(B, DNFB, Level),
        distribute(DNFA, DNFB, DNFAB).
% Beachte: die Optimierung ist nur in der "au"serten Konjunktion g"ultig.
% Die Klausel `p(X) <- X=1 or X=2' kann nicht optimiert werden.
disjunctive_nf(A=B, DNF, yes) :- !,
	(   unify(A,B) ->
		DNF = true
	;	DNF = fail
	).
disjunctive_nf(A, A, _Level).

distribute(A, or(B,C), ABC) :- !,
        distribute(A, B, AB),
        distribute(A, C, AC),
	disjunction(AB, AC, ABC).
distribute(or(A,B), C, ABC) :- !,
        distribute(A, C, AC),
        distribute(B, C, BC),
	disjunction(AC, BC, ABC).
distribute(A, B, AB) :-
	conjunction(A, B, AB).

disjunction(true, _B, true) :- !.
disjunction(_A, true, true) :- !.
disjunction(fail, B, B) :- !.
disjunction(A, fail, A) :- !.
disjunction(A, B, or(A,B)).

conjunction(true, B, B) :- !.
conjunction(A, true, A) :- !.
conjunction(fail, _B, fail) :- !.
conjunction(_A, fail, fail) :- !.
conjunction(A, B, (A,B)).

contains(Var1, Type1, Assumptions) :-
	(   var(Assumptions) ->
		Assumptions=[Var1:Type1|_]
	;	Assumptions=[Var2:Type2|Rest], 
		(   Var1==Var2 ->
			Type1=Type2
		;	contains(Var1, Type1, Rest)
		)
	).

check(Type1, Type2) :-
	( unify(Type1, Type2) ->
		true
	;	ddbs_error('*** cannot unify ~w with ~w~n', [Type1,Type2])
	).

%--------------------------- Aenderungsbearbeitung ----------------------------

evaluate_update(Update,Del,Ins) :-
        deletes(Update,Del),
        inserts(Update,Ins).

deletes(P,Del) :-
        user:update_proc(P,Body),
        !,
        deletes(Body,Del). 
deletes(+_A,[]) :- !.
deletes(-A,[A]) :- !.
deletes(P <- W,Del) :- !,
        findall(DelP,(user:W,deletes(P,DelP)),Dels),
        union(Dels,Del).
deletes(Ps,Del) :- !,
        findall(DelP,(member(P,Ps),deletes(P,DelP)),Dels),
        union(Dels,Del).

inserts(P,Ins) :-
        user:update_proc(P,Body),
        !,
        inserts(Body,Ins).
inserts(+A,[A]) :- !.
inserts(-_A,[]) :- !.
inserts(P <- W,Ins) :- !,
        findall(DelP,(user:W,inserts(P,DelP)),Dels),
        union(Dels,Ins).
inserts(Ps,Ins) :- !,
        findall(DelP,(member(P,Ps),inserts(P,DelP)),Dels),
        union(Dels,Ins).


print_update([],_Text).
print_update([ParPred|ParPreds],Text) :-
        pred_decl(_,Spec,ParPred),
        same_functor(Spec,Pred,_,Arity),
        parform_to_pred(Arity,ParPred,Pred),
        print(Text),
        print(Pred),
        nl,
        print_update(ParPreds,Text).

parform_to_pred(N,ParPred,Pred) :-
        (   N =:= 0 -> true
        ;       arg(N,ParPred,Arg),
                arg(N,Pred,Arg),
                N1 is N - 1,
                parform_to_pred(N1,ParPred,Pred)
        ).

%------------------------------ Fehlerbehandlung -------------------------------

ddbs_error(Formats, Args) :-
	format(Formats, Args), 
	assert(ddbs_error).

%------------------------------- Verschiedenes ---------------------------------

save :-
	retract_clauses,
	retract_dynamic,
	save_program('ddbs', ddbs).

rjustify(N, String) :-
	length(String, M),
	Spaces is N-M,
	tab(Spaces),
	format('~s~n',[String]).

forall(Goal, Condition) :-
	\+ (Goal, \+ Condition).

member(A, [A|_X]).
member(A, [_B|X]) :-
	member(A, X).

%%% stolen and modified from the library `freevars.pl'.

free_variables(Term, Bound, Vars) :-
	free_variables(Term, Bound, [], Vars).

free_variables(Term, Bound, Vars0, Vars) :-
	(   nonvar(Term) ->
	    free_variables_1(Term, Bound, Vars0, Vars)
	;   term_is_free_of(Bound, Term),
	    list_is_free_of(Vars0, Term)
	->  Vars = [Term|Vars0]
	;   Vars = Vars0
	).

free_variables_1(Term:_Type, Bound, Vars0, Vars) :-
	free_variables(Term, Bound, Vars0, Vars).
free_variables_1(Form, Bound, Vars0, Vars) :-
	comprehension(Form, Head, Body, _), !,
	free_variables(Head, [], [], Free),
	free_variables(Body, Free^Bound, Vars0, Vars).
free_variables_1(Form, Bound, Vars0, Vars) :-
	set_expression(Form, Template, Generator, Set), !,
	free_variables(Template, [], [], Free),
        free_variables(Generator, Free^Bound, Vars0, Vars1),
        free_variables(Set, Bound, Vars1, Vars).
free_variables_1(Form, Bound, Vars0, Vars) :-
	quan_decl(Form, _, _, Arity), !,
	arg(1, Form, Local),
	free_variables_arg(2, Arity, Form, Local^Bound, Vars0, Vars).
free_variables_1(Form, Bound, Vars0, Vars) :-
	functor(Form, _, Arity),
	free_variables_arg(1, Arity, Form, Bound, Vars0, Vars).

free_variables_arg(N, Arity, Term, Bound, Vars0, Vars) :-
        (   N>Arity -> Vars0=Vars
    	;	arg(N, Term, Arg),
		free_variables(Arg, Bound, Vars0, Vars1),
		M is N + 1,
		free_variables_arg(M, Arity, Term, Bound, Vars1, Vars)
	).

comprehension([Compr], Head, Body, listof) :-
	nonvar(Compr),
	Compr = (Head<=Body).
comprehension({{Head|Body}}, Head, Body, listof).
comprehension([Compr], Head, Body, permof) :-
	nonvar(Compr),
	Compr = (Head<-Body).
comprehension({Head|Body}, Head, Body, permof).

set_expression(setof(Template,Generator,Set), Template, Generator, Set).
set_expression(permof(Template,Generator,Set), Template, Generator, Set).
set_expression(bagof(Template,Generator,Set), Template, Generator, Set).
set_expression(listof(Template,Generator,Set), Template, Generator, Set).

%%% stolen from the library `freevars.pl'.

term_is_free_of(Term, Var) :-
        (   var(Term) ->
            Term \== Var
        ;   functor(Term, _, N),
            term_is_free_of(N, Term, Var)
        ).
term_is_free_of(N, Term, Var) :-
        (   N =:= 0 -> true
        ;   arg(N, Term, Arg),
            term_is_free_of(Arg, Var),
            M is N-1,
            term_is_free_of(M, Term, Var)
        ).

%%% stolen from the library `freevars.pl'.

list_is_free_of([], _).
list_is_free_of([Head|Tail], Var) :-
        Head \== Var,
        list_is_free_of(Tail, Var).

list_to_conjunction([], true).
list_to_conjunction([A|Atoms], and(A,Conj)) :-
        list_to_conjunction(Atoms, Conj).
